# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import pickle
import sys

# from msilib.schema import Error
from pathlib import Path
from shutil import copyfile
from typing import Literal, Optional, Union

import pandas as pd
from sapientml.model import GeneratedModel
from sapientml.suggestion import SapientMLSuggestion

from .macros import Metric
from .params import Dataset, Task
from .util.logging import setup_logger

if sys.version_info.minor <= 9:
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

logger = setup_logger()

DEFAULT_OUTPUT_DIR = "./outputs"


def _check_stratification(
    full_dataset,
    target_columns,
    task_type,
    adaptation_metric,
    split_stratification,
    stratify_threshold,
):
    base_threshold = 0.2
    temp_res = full_dataset[target_columns[0]].value_counts()
    min_value = temp_res.min()
    min_value_ratio = min_value / temp_res.sum()
    ratio_threshold = base_threshold / len(temp_res)
    if task_type == "classification":
        if min_value < stratify_threshold:
            if adaptation_metric == "LogLoss":
                raise ValueError(
                    f"Target value {temp_res.idxmin()} appears only {min_value} times, the input data cannot be split for LogLoss metric. Please review the target column or set another adaptation_metric"
                )
            elif split_stratification:
                raise ValueError(
                    f"split_stratification=True, but the input data cannot be split using stratification because target value {temp_res.idxmin()} appears only {min_value} times. Please set split_stratification as False"
                )
            else:
                return False
        elif min_value_ratio < ratio_threshold:
            if split_stratification is None:
                return True
            elif not split_stratification:
                logger.warning(
                    "split_stratification is set False, but setting True is recommended because the target is imbalance."
                )
            return split_stratification
        else:
            if split_stratification is None:
                return False
            return split_stratification
    else:
        return False


class SapientML:
    def __init__(
        self,
        # Dataset
        target_columns: list[str],
        # Task
        task_type: Optional[Literal["classification", "regression"]] = None,
        adaptation_metric: Optional[str] = None,
        split_method: Literal["random", "time", "group"] = "random",
        split_seed: int = 17,
        split_train_size: float = 0.75,
        split_column_name: Optional[str] = None,
        time_split_num: int = 5,
        time_split_index: int = 4,
        split_stratification: Optional[bool] = None,
        model_type: str = "sapientml",
        **kwargs,
    ):
        """
        The constructor of SapientML.

        You can pass all the keyword arguments for configurations of PipelineGenerators and CodeBlockGenerators from plugins.

        Parameters
        ----------
        target_columns: list[str]
            Names of target columns.
        task_type: 'classification', 'regression' or None
            Specify task type classification, regression.
        adaptation_metric: str
            Metric for evaluation.
            Classification: 'f1', 'auc', 'ROC_AUC', 'accuracy', 'Gini', 'LogLoss',
            'MCC'(Matthews correlation coefficient), 'QWK'(Quadratic weighted kappa).
            Regression: 'r2', 'RMSLE', 'RMSE', 'MAE'.
        split_method: 'random', 'time', or 'group'
            Method of train-test split.
            'random' uses random split.
            'time' requires 'split_column_name'.
            This sorts the data rows based on the column, and then splits data.
            'group' requires 'split_column_name'.
            This splits the data so that rows with the same value of 'split_column_name'
            are not placed in both training and test data.
        split_seed: int
            Random seed for train-test split.
            Ignored when split_method='time'.
        split_train_size: float
            The ratio of training size to input data.
            Ignored when split_method='time'.
        split_column_name: str
            Name of the column used to split.
            Ignored when split_method='random'
        time_split_num: int
            Passed to TimeSeriesSplit's n_splits.
            Valid only when split_method='time'.
        time_split_index: int
            The index of the split from TimeSeriesSplit.
            Valid only when split_method='time'.
        split_stratification: bool
            To perform stratification in train-test split.
            Valid only when task_type='classification'.

        """
        self.task = Task(
            target_columns=target_columns,
            task_type=task_type,
            split_method=split_method,
            split_seed=split_seed,
            split_train_size=split_train_size,
            split_column_name=split_column_name,
            time_split_num=time_split_num,
            time_split_index=time_split_index,
            adaptation_metric=adaptation_metric,
            split_stratification=split_stratification,
        )
        eps_generator = entry_points(group="sapientml.pipeline_generator")
        eps_config = entry_points(group="sapientml.config")
        if eps_generator[model_type] and eps_config[model_type]:
            self._Generator = eps_generator[model_type].load()
            self._Config = eps_config[model_type].load()
        else:
            raise ValueError(f"Model '{model_type}' is invalid.")

        self.model_type = model_type
        self.generator = self._Generator(**kwargs)
        self.config = self.generator.config
        self.config.post_init()

    @staticmethod
    def from_pretrained(model):
        """
        The factory method of SapientML from a pretrained model built by source code previously generated by SapientML.

        `model` must be either pickle filename, pickle bytes-like object, or deserialized object


        Parameters
        ----------
        model: str, bytes-like object, or GeneratedModel
            A pretrained model built by source code previously generated by SapientML.

        Returns
        -------
        sml: SapientML
            a new SapientML instance loaded from the pretrained model.
        """
        try:
            if isinstance(model, GeneratedModel):
                pass
            else:
                if isinstance(model, str):
                    with open(model, "rb") as f:
                        model = pickle.load(f)
                else:
                    model = pickle.loads(model)
                if not isinstance(model, GeneratedModel):
                    raise RuntimeError("model is not an instance of GeneratedModel")
        except Exception as e:
            raise ValueError(
                "model must be either pickle filename, pickle bytes-like object, or deserialized object"
            ) from e
        sml = SapientML(**model.params)
        sml.model = model
        return sml

    def fit(
        self,
        training_data: Union[pd.DataFrame, str],
        validation_data: Optional[Union[pd.DataFrame, str]] = None,
        test_data: Optional[Union[pd.DataFrame, str]] = None,
        save_datasets_format: Literal["csv", "pickle"] = "pickle",
        csv_encoding: Literal["UTF-8", "SJIS"] = "UTF-8",
        csv_delimiter: str = ",",
        ignore_columns: Optional[list[str]] = None,
        output_dir: str = DEFAULT_OUTPUT_DIR,
        codegen_only: bool = False,
    ):
        """
        Generate ML scripts for input data.

        Parameters
        ----------
        training_data: pandas.DataFrame or str
            Training dataframe.
            When str, this is regarded as a file path.
        validation_data: pandas.DataFrame, str or None
            Validation dataframe.
            When str, this is regarded as file paths.
            When None, validation data is extracted from training data by split.
        test_data: pandas.DataFrame, str, or None
            Test dataframes.
            When str, they are regarded as file paths.
            When None, test data is extracted from training data by split.
        save_datasets_format: 'csv' or 'pickle'
            Data format when the input dataframes are written to files.
            Ignored when all inputs are specified as file path.
        csv_encoding: 'UTF-8' or 'SJIS'
            Encoding method when csv files are involved.
            Ignored when only pickle files are involved.
        csv_delimiter: str
            Delimiter to read csv files.
        ignore_columns: list[str]
            Column names which must not be used and must be dropped.
        output_dir: str
            Output directory.
        codegen_only: bool
            Do not conduct fit() of GeneratedModel if True.

        Returns
        -------
        self: SapientML
            SapientML object itself.
        """

        if ignore_columns is None:
            ignore_columns = []

        logger.info("Loading dataset...")

        self.output_dir = Path(output_dir).resolve()
        self.output_dir.mkdir(parents=True, exist_ok=True)

        # copy libs
        lib_path = self.output_dir / "lib"
        lib_path.mkdir(exist_ok=True)

        eps = entry_points(group="sapientml.export_modules")
        for ep in eps:
            if ep.name in [self.generator.__class__.__name__, "sample-dataset"]:
                for file in glob.glob(f"{ep.load().__path__[0]}/*.py"):
                    copyfile(file, lib_path / Path(file).name)

        self.dataset = Dataset(
            training_data=training_data,
            validation_data=validation_data,
            test_data=test_data,
            csv_encoding=csv_encoding,
            csv_delimiter=csv_delimiter,
            save_datasets_format=save_datasets_format,
            ignore_columns=ignore_columns,
            output_dir=self.output_dir,
        )

        if self.task.task_type is None:
            self.task.task_type = SapientMLSuggestion(
                self.task.target_columns, self.dataset.training_dataframe
            ).suggest_task()

        if self.task.adaptation_metric is None and self.task.task_type:
            if self.task.task_type == "regression":
                logger.warning("Metric is not specified. Use 'r2' by default.")
            else:
                logger.warning("Metric is not specified. Use 'f1' by default.")
            self.task.adaptation_metric = Metric.get_default_value(self.task.task_type)

        self.task.adaptation_metric = Metric.get(self.task.adaptation_metric)

        if not Metric.metric_match_task_type(self.task.adaptation_metric, self.task.task_type):
            logger.warning(f"{self.task.adaptation_metric} is not a metric for {self.task.task_type}")

        if self.task.task_type == "classification":
            is_multioutput = False
            if len(self.task.target_columns) > 1:
                is_multioutput = True
            for column in self.task.target_columns:
                if isinstance(training_data, pd.DataFrame) and len(training_data[column].unique()) > 2:
                    self.task.is_multiclass = True

            if is_multioutput:
                if self.task.is_multiclass and not Metric.metric_support_multiclass_multioutput(
                    self.task.adaptation_metric
                ):
                    logger.warning(
                        f"{self.task.adaptation_metric} does not support Multiclass-Multioutput classification. Execution of candidate script raises an exception."
                    )
                elif not Metric.metric_support_multioutput(self.task.adaptation_metric):
                    logger.warning(
                        f"{self.task.adaptation_metric} does not support Multioutput (Multilabel) classification. Execution of candidate script raises an exception."
                    )

        training_dataframe, validation_dataframe, test_dataframe = self.dataset.get_dataframes()
        stratify_threshold = 3
        if validation_dataframe is not None:
            stratify_threshold -= 1
        if test_dataframe is not None:
            stratify_threshold -= 1
        if len(self.task.target_columns) == 1:
            self.task.split_stratification = _check_stratification(
                training_dataframe,
                self.task.target_columns,
                self.task.task_type,
                self.task.adaptation_metric,
                self.task.split_stratification,
                stratify_threshold,
            )
        elif self.task.task_type == "classification" and self.task.split_stratification:
            raise ValueError("Stratification for multiple target columns is not supported.")

        self.generator.generate_pipeline(self.dataset, self.task)
        self.dataset.reload()
        self.generator.save(self.output_dir)

        self.params = {"model_type": self.model_type}
        self.params.update(self.task.model_dump())
        self.params.update(self.config.model_dump())
        self.params.update(self.generator.datastore.config.model_dump())
        self.params.update(self.generator.preprocess.config.model_dump())

        self.model = GeneratedModel(
            input_dir=self.output_dir,
            save_datasets_format=save_datasets_format,
            csv_encoding=csv_encoding,
            csv_delimiter=csv_delimiter,
            timeout=self.config.timeout_for_test,
            params=self.params,
        )

        if not codegen_only:
            self.model.fit(pd.concat([training_dataframe, validation_dataframe]))

        logger.info("Done.")

        return self

    def predict(
        self,
        test_data: pd.DataFrame,
    ):
        """
        Predicts the output of the test_data.

        Parameters
        ---------
        test_data: pd.DataFrame
            Dataframe used for predicting the result.

        Returns
        -------
        result : pd.DataFrame
            It returns the prediction_result.csv result in dataframe format.

        """
        logger.info("Predicting by built model...")
        return self.model.predict(test_data)



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import os
import platform
import sys
import time
from pathlib import Path
from typing import Optional

import nest_asyncio

from .params import CancellationToken, Code, RunningResult
from .util.logging import setup_logger

logger = setup_logger()


def run(
    file_path: str, timeout: int, cancel: Optional[CancellationToken] = None, cwd: Optional[str] = None
) -> RunningResult:
    """Executing run() function based on operating system.

    Parameters
    ----------
    filepath : str
        Path of the file executed.
    timeout : int
        Timeout for the execution.
    cancel : CancellationToken, optional
        Object for cancellation.
    cwd : str, optional
        Working directory.

    Returns
    -------
    result : RunningResult

    """

    if platform.system() == "Windows":
        encoding = "cp932"  # noqa
        replace_newline = "\r"  # noqa
        loop = asyncio.ProactorEventLoop()  # noqa
    else:
        encoding = "utf-8"
        replace_newline = ""
        loop = asyncio.new_event_loop()

    asyncio.set_event_loop(loop)
    nest_asyncio.apply(loop)

    async def _run() -> RunningResult:
        start_time = time.time()

        process = await asyncio.create_subprocess_exec(
            sys.executable,
            file_path,
            cwd=cwd or os.path.dirname(file_path),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        endtime = time.time() + timeout if timeout > 0 else None

        async def _read_stream(stream, encoding, replace_newline):
            lines = []
            while True:
                line = await stream.readline()
                if line:
                    lines.append(line.decode(encoding).replace(replace_newline, ""))
                if stream.at_eof():
                    break
                await asyncio.sleep(0.1)
            return lines

        async def _wait_timeout_or_cancel(proc, endtime, cancel):
            interrupted_reason = None
            returncode = None

            while True:
                if proc.stdout.at_eof() and proc.stderr.at_eof():
                    break

                if endtime is not None and time.time() > endtime:
                    returncode = -9
                    interrupted_reason = "Timeout"
                    print("Terminating due to timeout")
                    proc.kill()
                    break
                if cancel is not None and cancel.is_triggered:
                    returncode = -9  # noqa
                    interrupted_reason = "Cancelled by user"  # noqa
                    print("Terminating due to cancellation")  # noqa
                    proc.kill()  # noqa
                    break  # noqa
                await asyncio.sleep(1)
            return returncode, interrupted_reason

        results = await asyncio.gather(
            _read_stream(process.stdout, encoding, replace_newline),
            _read_stream(process.stderr, encoding, replace_newline),
            _wait_timeout_or_cancel(process, endtime, cancel),
        )

        await process.wait()

        stdout_lines = results[0]
        stderr_lines = results[1]
        returncode, interrupted_reason = results[2]

        if not returncode:
            returncode = process.returncode

        if interrupted_reason is not None:
            output = ""
            error = interrupted_reason
        else:
            output = "".join(stdout_lines)
            error = "".join(stderr_lines)

        result = RunningResult(
            output=output,
            error=error,
            returncode=returncode,
            time=int(round(time.time() - start_time)),
        )
        return result

    try:
        result = loop.run_until_complete(_run())
    finally:
        loop.close()
        asyncio.set_event_loop(asyncio.new_event_loop())

    return result


class PipelineExecutor:
    """PipelineExecutor class for executing the generated pipelines."""

    def __init__(
        self,
    ):
        pass

    def execute(
        self,
        pipeline_list: list[Code],
        initial_timeout: int,
        output_dir: Path,
        cancel: Optional[CancellationToken],
    ) -> list[tuple[Code, RunningResult]]:
        """Executes the generated pipelines.

        Parameters
        ----------
        pipeline_list: list[Code]
            List of generated pipeline code.
        initial_timeout: int
            Timeout for the execution.
        output_dir: Path
            Output directory to store the results.
        cancel : CancellationToken, optional
            Object for cancellation.
        Returns
        -------
        candidate_scripts: list[tuple[Code, RunningResult]]
            It stores both the results and the code in list of tuples format.

        """
        candidate_scripts: list[tuple[Code, RunningResult]] = []

        for index, pipeline in enumerate(pipeline_list, start=1):
            script_name = f"{index}_script.py"
            script_path = (output_dir / script_name).absolute().as_posix()
            with open(script_path, "w", encoding="utf-8") as f:
                f.write(pipeline.validation)

            logger.info(f"Running script ({index}/{len(pipeline_list)})...")
            running_result = run(script_path, initial_timeout, cancel)
            candidate_scripts.append((pipeline, running_result))
            reason = ""
            error_message = running_result.error.strip().split("\n")
            if running_result.returncode != 0:
                if running_result.returncode == 124:
                    # Status code 124 means timeout on linux timeout command
                    reason = "Timeout"
                elif error_message and error_message[-1]:
                    # Show the stack trace so users can get full error info
                    reason = "\n".join(error_message)
                else:
                    # Unknown error
                    reason = f"Status code: {running_result.returncode}"
                logger.warning(f"Failed to run a pipeline '{script_name}': {reason}")

        return candidate_scripts



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Tuple, Union

from .params import Code, Dataset, Task


class CodeBlockGenerator(ABC):
    def __init__(self, **kwargs):  # pragma: no cover
        pass

    @abstractmethod
    def generate_code(self, dataset: Dataset, task: Task) -> Tuple[Dataset, Code]:  # pragma: no cover
        pass


class PipelineGenerator(ABC):
    @abstractmethod
    def generate_pipeline(self, dataset: Dataset, task: Task):  # pragma: no cover
        pass

    @abstractmethod
    def save(self, output_dir: Union[Path, str]):  # pragma: no cover
        pass

    @abstractmethod
    def add_explanation(self):  # pragma: no cover
        pass



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from pathlib import Path
from typing import Annotated, List, Literal, Optional, Union

import numpy as np
import pandas as pd
from pydantic import BaseModel, Field, StringConstraints, field_validator

from .macros import Metric
from .util.logging import setup_logger

MAX_NUM_OF_COLUMNS = 10000000
MAX_PATH_LENGTH = 1000
MAX_SEED = 2**32 - 1
MAX_COLUMN_NAME_LENGTH = 1000
MAX_TIME_SPLIT_NUM = 10000
MAX_TIME_SPLIT_INDEX = 10000
DEFAULT_OUTPUT_DIR = "./outputs"
INITIAL_TIMEOUT = 600

MAX_STR_LENGTH = 1000
MAX_INT = 1000000000

Int = Annotated[int, Field(strict=False, gt=-MAX_INT, lt=MAX_INT)]
PositiveInt = Annotated[int, Field(strict=False, gt=0, lt=MAX_INT)]
NonNegativeInt = Annotated[int, Field(strict=False, ge=0, lt=MAX_INT)]
Float = Annotated[float, Field(strict=False, gt=-MAX_INT, lt=MAX_INT)]
NonNegativeFloat = Annotated[float, Field(strict=False, ge=0, lt=MAX_INT)]
Fraction = Annotated[float, Field(strict=False, ge=0, le=1)]
String = Annotated[str, StringConstraints(max_length=MAX_STR_LENGTH)]

ColumnName = Annotated[str, StringConstraints(min_length=1, max_length=MAX_COLUMN_NAME_LENGTH - 1)]
Seed = Annotated[int, Field(strict=False, ge=0, le=MAX_SEED)]

logger = setup_logger()


def _read_file(filepath: str, csv_encoding: str, csv_delimiter: str) -> pd.DataFrame:
    if filepath.endswith(".pkl"):
        res_df = pd.read_pickle(filepath)
    else:
        res_df = pd.read_csv(filepath, encoding=csv_encoding, delimiter=csv_delimiter)
    return res_df


def save_file(dataframe: pd.DataFrame, filepath: str, csv_encoding: str, csv_delimiter: str) -> None:
    """Saving dataframe to pickle or csv files

    Parameters
    ----------
    dataframe : pd.DataFrame
        The data to output
    filepath : str
        The path of output file
    csv_encoding : 'UTF-8' or 'SJIS'
        Encoding method when csv files are involved.
        Ignored when only pickle files are involved.
    csv_delimiter : str
        Delimiter to read csv files
    """
    if filepath.endswith(".pkl"):
        dataframe.to_pickle(filepath)
    else:
        dataframe.to_csv(filepath, encoding=csv_encoding, sep=csv_delimiter, index=False)


def _is_strnum_column(c):
    c2 = c.loc[c.notnull()]
    c2 = pd.to_numeric(c2, errors="coerce")
    ratio = c2.notnull().sum() / c2.shape[0]
    return ratio > 0.9


def _confirm_mixed_type(df: pd.DataFrame) -> list[str]:
    mixed_df_cols = []
    for df_col in df.columns:
        df_per_col = df[df_col].replace([-np.inf, np.inf], np.nan)
        df_per_col2 = df_per_col[df_per_col.notnull()]
        types_per_col = pd.api.types.infer_dtype(df_per_col2)

        is_strnum = _is_strnum_column(df[df_col])

        if types_per_col == "mixed" or types_per_col == "mixed-integer" or (types_per_col == "string" and is_strnum):
            mixed_df_cols.append(df_col)

    return mixed_df_cols


def _is_date_column(c):
    c2 = c.loc[c.notnull()]
    if c2.shape[0] > 1000:
        c2 = c2.sample(1000, random_state=17)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        c2 = pd.to_datetime(c2, errors="coerce")
    ratio = c2.notnull().sum() / c2.shape[0]
    return ratio > 0.8


class CancellationToken(BaseModel):
    """CancellationToken class.

    Attributes
    ----------
    is_triggered : bool

    """

    is_triggered: bool = False


class Code(BaseModel):
    """Code class.

    Attributes
    ----------
    validation : str
    test : str
    train : str
    predict : str

    """

    validation: str = ""
    test: str = ""
    train: str = ""
    predict: str = ""

    def __add__(self, other):
        return Code(
            validation=self.validation + other.validation,
            test=self.test + other.test,
            train=self.train + other.train,
            predict=self.predict + other.predict,
        )


class RunningResult(BaseModel):
    """RunningResult class.

    Attributes
    ----------
    output : str
    error : str
    returncode : int
    time : int

    """

    output: str
    error: str
    returncode: int
    time: NonNegativeInt


class PipelineResult(BaseModel):
    """PipelineResult class.

    Attributes
    ----------
    score : float, optional
    metric : str, optional
    best_params : dict, optional

    """

    score: Optional[float]
    metric: Optional[str]
    best_params: Optional[dict]


class Task(BaseModel):
    """Task class.

    Attributes
    ----------
    target_columns : list[str]
        Names of target columns
    task_type : 'classification' or 'regression'
        Specify classification or regression.
    ignore_columns: list[str]
        list of columns which are not necessary or ignored.
    split_method : 'random', 'time', or 'group'
        Method of train-test split.
        'random' uses random split.
        'time' requires 'split_column_name'.
        This sorts the data rows based on the column, and then splits data.
        'group' requires 'split_column_name'.
        This split the data so as not to split rows with the same value of 'split_column_name'
        into train and test data.
        Currently, this option is not valid in the hyperparameter tuning.
        Don't set time or group when hyperparameter_tuning=True.
    split_seed : int
        Random seed for train-test split.
        Ignored when split_method='time'.
    split_train_size : float
        The ratio of training size to input data.
        Ignored when split_method='time'.
    split_column_name : str
        Name of the column used to split.
        Ignored when split_method='random'
    time_split_num : int
        Passed to TimeSeriesSplit's n_splits.
        Valid only when split_method='time'.
    time_split_index : int
        The index of the split from TimeSeriesSplit.
        Valid only when split_method='time'.
    adaptation_metric : str
        Metric for evaluation.
        Classification: 'f1', 'auc', 'ROC_AUC', 'accuracy', 'Gini', 'LogLoss',
        'MCC'(Matthews correlation coefficient), 'QWK'(Quadratic weighted kappa).
        Regression: 'r2', 'RMSLE', 'RMSE', 'MAE'.
    is_multiclass: bool
        To check whether it is a multiclass or not.
    split_stratification : bool
        To perform stratification in train-test split.

    """

    target_columns: list[ColumnName]
    task_type: Optional[Literal["classification", "regression"]]
    split_method: Literal["random", "time", "group"]
    split_seed: Seed
    split_train_size: Float
    split_column_name: Optional[ColumnName]
    time_split_num: PositiveInt
    time_split_index: NonNegativeInt
    adaptation_metric: Optional[String]
    is_multiclass: bool = False
    split_stratification: Optional[bool] = None

    @field_validator("target_columns")
    def _check_num_of_column_names(cls, v):
        if v is None:
            return v
        if len(v.keys() if isinstance(v, dict) else v) >= MAX_NUM_OF_COLUMNS:
            raise ValueError(f"The number of columns must be smaller than {MAX_NUM_OF_COLUMNS}")
        return v

    @field_validator("target_columns")
    def _check_columns_is_not_empty(cls, v):
        if v is None:
            return v
        if len(v) == 0:
            raise ValueError("Target columns are empty.")
        return v

    @field_validator("split_train_size")
    def _check_split_train_size(cls, v):
        if v <= 0 or 1 <= v:
            raise ValueError(f"{v} is out of (0, 1)")
        return v

    @field_validator("time_split_num")
    def _check_time_split_num(cls, v):
        if v < 1 or MAX_TIME_SPLIT_NUM < v:
            raise ValueError(f"{v} is out of [1, {MAX_TIME_SPLIT_NUM}]")
        return v

    @field_validator("time_split_index")
    def _check_time_split_index(cls, v):
        if v < 0 or MAX_TIME_SPLIT_INDEX < v:
            raise ValueError(f"{v} is out of [1, {MAX_TIME_SPLIT_INDEX}].")
        return v

    @field_validator("adaptation_metric")
    def _check_metric(cls, v):
        if v is None:
            return v
        try:
            Metric.get(v)
        except NotImplementedError:
            raise ValueError(f"'{v}' is invalid as a metric.")
        return v


class Config(BaseModel):
    """
    Configuration arguments for sapientml.generator.CodeBlockGenerator
    and/or sapientml.generator.PipelineGenerator.

    Attributes
    ----------
    initial_timeout : int
        Timelimit to execute each generated script.
        Ignored when hyperparameter_tuning=True and hyperparameter_tuning_timeout is set.
    timeout_for_test : int
        Timelimit to execute test script (final_script) and Visualization.
    cancel : CancellationToken, optional
        Object to interrupt evaluations.
    project_name : str, optional
        Project name.
    debug : bool
        Debug mode or not.
    """

    initial_timeout: NonNegativeInt = INITIAL_TIMEOUT
    timeout_for_test: NonNegativeInt = 0
    cancel: Optional[CancellationToken] = None
    project_name: Optional[String] = None
    debug: bool = False


class Dataset:
    def __init__(
        self,
        training_data: Union[pd.DataFrame, str],
        validation_data: Optional[Union[pd.DataFrame, str]] = None,
        test_data: Optional[Union[pd.DataFrame, str]] = None,
        csv_encoding: Literal["UTF-8", "SJIS"] = "UTF-8",
        csv_delimiter: str = ",",
        save_datasets_format: Literal["csv", "pickle"] = "pickle",
        ignore_columns: Optional[List[str]] = None,
        output_dir: Path = Path(DEFAULT_OUTPUT_DIR),
    ):
        """
        Checking/Preparing the dataset.

        Parameters
        ----------
        training_data: pandas.DataFrame or str
            Training dataframe.
            When str, this is regarded as a file path.
        validation_data: pandas.DataFrame, str or None
            Validation dataframe.
            When str, this is regarded as file paths.
            When None, validation data is extracted from training data by split.
        test_data: pandas.DataFrame, str, or None
            Test dataframes.
            When str, they are regarded as file paths.
            When None, test data is extracted from training data by split.
        csv_encoding: 'UTF-8' or 'SJIS'
            Encoding method when csv files are involved.
            Ignored when only pickle files are involved.
        csv_delimiter: str
            Delimiter to read csv files
        save_datasets_format: 'csv' or 'pickle'
            Data format when the input dataframes are written to files.
            Ignored when all inputs are specified as file path.
        ignore_columns: list[str]
            Column names which must not be used and must be dropped.
        output_dir: str
            Output dir

        """
        self.ignore_columns = [] if ignore_columns is None else ignore_columns
        self.csv_encoding = csv_encoding
        self.csv_delimiter = csv_delimiter
        self.save_datasets_format = save_datasets_format
        self.output_dir = output_dir

        if isinstance(training_data, str):
            self.training_dataframe = _read_file(training_data, csv_encoding, csv_delimiter)
            self.training_data_path = training_data
        elif isinstance(training_data, pd.DataFrame):
            self.training_dataframe = training_data.copy()
            filename = "training." + ("pkl" if save_datasets_format == "pickle" else "csv")
            self.training_data_path = str(self.output_dir / filename)
            save_file(self.training_dataframe, self.training_data_path, csv_encoding, csv_delimiter)

        # NOTE: self.validation_data and self.test_data can be None
        if validation_data is not None and test_data is None:
            raise ValueError(
                "test_data must not be None when validation_data is specified. test_data should be specified instead of validation_data."
            )
        if isinstance(validation_data, str):
            self.validation_dataframe = _read_file(validation_data, csv_encoding, csv_delimiter)
            self.validation_data_path = validation_data
        elif isinstance(validation_data, pd.DataFrame):
            self.validation_dataframe = validation_data.copy()
            filename = "validation." + ("pkl" if save_datasets_format == "pickle" else "csv")
            self.validation_data_path = str(self.output_dir / filename)
            save_file(self.validation_dataframe, self.validation_data_path, csv_encoding, csv_delimiter)
        else:
            self.validation_dataframe = None
            self.validation_data_path = None

        if isinstance(test_data, str):
            self.test_dataframe = _read_file(test_data, csv_encoding, csv_delimiter)
            self.test_data_path = test_data
        elif isinstance(test_data, pd.DataFrame):
            self.test_dataframe = test_data.copy()
            filename = "test." + ("pkl" if save_datasets_format == "pickle" else "csv")
            self.test_data_path = str(self.output_dir / filename)
            save_file(self.test_dataframe, self.test_data_path, csv_encoding, csv_delimiter)
        else:
            self.test_dataframe = None
            self.test_data_path = None

    def reload(self):
        self.training_dataframe = _read_file(self.training_data_path, self.csv_encoding, self.csv_delimiter)
        if self.validation_data_path:
            self.validation_dataframe = _read_file(self.validation_data_path, self.csv_encoding, self.csv_delimiter)
        if self.test_data_path:
            self.test_dataframe = _read_file(self.test_data_path, self.csv_encoding, self.csv_delimiter)

    def check_dataframes(
        self,
        target_columns: List[str],
    ):
        """Checking dataframes method.

        Parameters
        ----------
        target_columns : List[str]
            Names of target columns.

        """
        # 1. Check status of each dataframe
        self._check_single_dataframe(self.training_dataframe, target_columns, "train")
        if self.validation_dataframe is not None:
            self._check_single_dataframe(self.validation_dataframe, target_columns, "validation")
        if self.test_dataframe is not None:
            self._check_single_dataframe(self.test_dataframe, target_columns, "test")

        def format_warning(inconsistent_df_cols, target_data_name1, target_data_name2):
            for inconsistent_df_col in inconsistent_df_cols:
                logger.warning(
                    f"types of {inconsistent_df_col} are inconsistent between {target_data_name1} and {target_data_name2} dataframes."
                )

        # 2. Check whether column names and orders are consistency between two dataframes
        if self.validation_dataframe is not None:
            if not self._is_same(
                self.training_dataframe.drop(self.ignore_columns, axis=1, errors="ignore"),
                self.validation_dataframe.drop(self.ignore_columns, axis=1, errors="ignore"),
            ):
                raise ValueError(
                    "Train/validation dataframes have different column names or its order. It must be same among the dataframes."
                )
        if self.test_dataframe is not None:
            if not self._is_same(
                self.training_dataframe.drop(target_columns, axis=1).drop(self.ignore_columns, axis=1, errors="ignore"),
                self.test_dataframe.drop(target_columns, axis=1, errors="ignore").drop(
                    self.ignore_columns, axis=1, errors="ignore"
                ),
            ):
                raise ValueError(
                    "Train/test dataframes have different column names or its order. It must be same among the dataframes."
                )

        # 3. Check whether types are consistency between two dataframes
        if self.validation_dataframe is not None:
            format_warning(
                self._confirm_consistent_type(self.training_dataframe, self.validation_dataframe),
                "train",
                "validation",
            )
        if self.test_dataframe is not None:
            format_warning(
                self._confirm_consistent_type(self.training_dataframe, self.test_dataframe),
                "train",
                "test",
            )

    def _check_single_dataframe(self, df: pd.DataFrame, target_columns: List[str], target_data_name: str) -> bool:
        # 0. Target existence
        # test data doesn't have to have target columns because there is the case where only prediction is needed.
        if (target_data_name != "test") and (not set(target_columns).issubset(set(df.columns))):
            raise ValueError(f"{target_data_name} doesn't have target columns")

        # 1. Check NaN in target column
        if (target_data_name == "train") and not self._isnot_nan(df, target_columns, 10):
            raise ValueError(
                f"target column of {target_data_name} dataframe has more than 10 rows including NaN or Inf."
            )
        elif (
            (target_data_name != "train")
            and set(target_columns).issubset(set(df.columns))
            and not self._isnot_nan(df, target_columns, 0)
        ):
            raise ValueError(f"target column of {target_data_name} dataframe has NaN or Inf.")

        # 2. Check whether index names are unique
        if not self._is_index_unique(df):
            raise ValueError(f"Index names of {target_data_name} dataframe are not unique.")

        # 3. Check whether column names are unique
        if not self._is_columns_unique(df):
            logger.warning(f"Column names of {target_data_name} dataframe are not unique.")

        # 4. Check strnum mixed type
        mixed_df_cols = _confirm_mixed_type(df)
        for mixed_df_col in mixed_df_cols:
            logger.warning(f"{mixed_df_col} would have mixed type in {target_data_name} dataframe.")

        # 5. Check whether datetime columns are correct value
        incorrect_info = self._confirm_date_type(df)
        incorrect_cols = list(incorrect_info.keys())
        incorrect_summary = []
        num_shown_for_one_column = 3
        for col, dt_column in incorrect_info.items():
            for index, value in dt_column.iloc[:num_shown_for_one_column].items():
                incorrect_summary.append((col, index, value))
        if incorrect_summary:
            logger.warning(
                f"The following columns would have incorrect datetime values in {target_data_name} dataframe: {incorrect_cols}"
            )
            logger.warning(f"For example, (column_name, index, value) = {incorrect_summary}")

        return True

    def _isnot_nan(self, df: pd.DataFrame, target_columns: list[str], limit: int) -> bool:
        return not (df[target_columns].isin([np.inf, -np.inf, np.nan]).any(axis=1).sum() > limit)

    def _is_columns_unique(self, df: pd.DataFrame) -> bool:
        if df.columns.is_unique:
            return True
        return False

    def _is_index_unique(self, df: pd.DataFrame) -> bool:
        if df.index.is_unique:
            return True
        return False

    def _confirm_consistent_type(self, df1: pd.DataFrame, df2: pd.DataFrame) -> list[str]:
        inconsistent_df_cols = []
        for df1_col in df1.columns:
            df1_col_type = pd.api.types.infer_dtype(df1[df1_col], skipna=True)
            if df1_col not in df2:
                continue
            df2_col_type = pd.api.types.infer_dtype(df2[df1_col], skipna=True)

            if df1_col_type != df2_col_type:
                inconsistent_df_cols.append(df1_col)

        return inconsistent_df_cols

    def _confirm_date_type(self, df: pd.DataFrame) -> dict:
        incorrect_info = {}
        # If only _is_date_column, all integer columns can be date columns.
        object_columns = df.select_dtypes("object").columns
        date_col = [col for col in object_columns if _is_date_column(df[col])]
        for col in date_col:
            org_column = df.loc[df[col].notnull(), col]
            dt_column = pd.to_datetime(org_column, errors="coerce")
            incorrect_values = org_column.loc[dt_column.isnull()]
            if incorrect_values.shape[0] > 0:
                incorrect_info[col] = incorrect_values
        return incorrect_info

    def _is_same(self, df1, df2):
        df1_columns = df1.columns
        df2_columns = df2.columns

        # The length of columns must be same
        if len(df1_columns) != len(df2_columns):
            return False

        # The item and order of columns must be same
        for i in range(len(df1_columns)):
            if df1_columns[i] != df2_columns[i]:
                return False
        return True

    def get_dataframes(self) -> tuple[pd.DataFrame, Optional[pd.DataFrame], Optional[pd.DataFrame]]:
        """get_dataframes method.

        Results
        ----------
        tuple
            returns traning, validation, test dataframes in tuple format.

        """
        return (self.training_dataframe, self.validation_dataframe, self.test_dataframe)



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import pandas as pd


class SapientMLSuggestion:
    def __init__(
        self,
        target_columns: List[str],
        dataframe: pd.DataFrame,
    ):
        """
        Suggest classification or regression for the input target_columns.

        Parameters
        ----------
        target_columns: list[str]
            Names of target columns.
        dataframe: pandas.DataFrame
            Input dataframe.
        """

        self.target_columns = target_columns
        self.dataframe = dataframe

    def suggest_task(self, th_val=30):
        """Suggest classification or regression.

        Parameters
        ----------
        th_val : int

        Returns
        -------
        str
            It returns task_suggestion as regression or classification.

        """
        task_suggestion_dict = {}
        for target_column in self.target_columns:
            unique_num = self.dataframe[target_column][self.dataframe[target_column].notna()].nunique()
            data = self.dataframe[target_column][self.dataframe[target_column].notna()]
            ratio = unique_num / data.count()

            if self.dataframe[target_column].dtypes == "object":
                task_params = "classification"
            else:
                if ratio < 0.05 and unique_num < th_val:
                    task_params = "classification"
                else:
                    task_params = "regression"
            task_suggestion_dict[target_column] = task_params

        count_classification = list(task_suggestion_dict.values()).count("classification")
        count_regression = list(task_suggestion_dict.values()).count("regression")

        if count_classification > count_regression:
            task_suggestion = "classification"
        else:
            task_suggestion = "regression"

        return task_suggestion
